/*

Copyright (c) 2005-2014, University of Oxford.
All rights reserved.

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of Oxford nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "Protocol.hpp"

#include <cassert>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <iterator>
#include <set>
#include <boost/pointer_cast.hpp>
#include <boost/foreach.hpp>
#include <boost/assign/list_of.hpp>

#include "ExecutableSupport.hpp"
#include "Warnings.hpp"
#include "PetscTools.hpp"

#include "ProtoHelperMacros.hpp"
#include "VectorStreaming.hpp"
#include "BacktraceException.hpp"
#include "ExceptionSet.hpp"
#include "DebugProto.hpp"
#include "ProtocolTimer.hpp"

// Typedefs for use with BOOST_FOREACH and std::maps
typedef std::pair<std::string, std::string> StringPair;
typedef std::pair<std::string, ProtocolPtr> StringProtoPair;


Protocol::Protocol()
    : mpInputs(new Environment(true)),
      mpLibrary(new Environment),
      mpModelStateCollection(new ModelStateCollection),
      mWritePng(false),
      mParalleliseLoops(false)
{
    mpLibrary->SetDelegateeEnvironment(mpInputs->GetAsDelegatee());
}


/**
 * Reset the provided outputs collection, in order to allow a protocol to be (re-)run.
 * @param rOutputs  the protocol outputs
 */
void ResetOutputs(std::map<std::string, EnvironmentPtr>& rOutputs)
{
    BOOST_FOREACH(StringEnvPair output_env, rOutputs)
    {
        output_env.second->Clear();
    }
    // Ensure the final outputs environment exists
    EnvironmentPtr p_proto_outputs(new Environment);
    rOutputs[""] = p_proto_outputs;
}


void Protocol::FinaliseSetup()
{
    // Check for duplicate plot titles (otherwise only the last would be written to file)
    std::set<std::string> plot_titles;
    BOOST_FOREACH(PlotSpecificationPtr p_plot_spec, mPlotSpecifications)
    {
        PROTO_ASSERT2(plot_titles.find(p_plot_spec->rGetTitle()) == plot_titles.end(),
                      "Plot title '" << p_plot_spec->rGetTitle() << "' has already been used.",
                      p_plot_spec->GetLocationInfo());
        plot_titles.insert(p_plot_spec->rGetTitle());
    }
    // Set default input definitions
    mpInputs->ExecuteStatements(mInputStatements);
    // Add delegatees for any imported libraries
    Environment& r_library = rGetLibrary();
    std::set<std::string> library_prefixes;
    BOOST_FOREACH(StringProtoPair import, mImports)
    {
        library_prefixes.insert(import.first);
        r_library.SetDelegateeEnvironment(import.second->rGetLibrary().GetAsDelegatee(), import.first);
    }
    // Set delegatees for simulations
    BOOST_FOREACH(boost::shared_ptr<AbstractSimulation> p_sim, mSimulations)
    {
        Environment& r_sim_env = p_sim->rGetEnvironment();
        // Don't change the delegatee for an imported simulation
        if (!r_sim_env.GetDelegateeEnvironment())
        {
            r_sim_env.SetDelegateeEnvironment(r_library.GetAsDelegatee());
        }
    }
    // Make the library delegate to (empty at present) prefixed simulation results environments,
    // and store pointers to those environments in mOutputs.
    BOOST_FOREACH(boost::shared_ptr<AbstractSimulation> p_sim, mSimulations)
    {
        const std::string prefix = p_sim->GetOutputsPrefix();
        if (!prefix.empty())
        {
            PROTO_ASSERT2(library_prefixes.find(prefix) == library_prefixes.end(),
                          "The simulation prefix '" << prefix << "' has already been used for a protocol import.",
                          p_sim->GetLocationInfo());
            PROTO_ASSERT2(!mOutputs[prefix],
                          "The simulation prefix '" << prefix << "' has been used for multiple simulations.",
                          p_sim->GetLocationInfo());
            r_library.SetDelegateeEnvironment(p_sim->GetResultsEnvironment(), prefix);
            mOutputs[prefix] = p_sim->GetResultsEnvironment();
        }
    }
}


void Protocol::InitialiseLibrary(bool reinit)
{
    if (reinit)
    {
        mpLibrary->Clear();
        mManifest.Clear();
    }
    // First initialise the libraries of imported protocols
    BOOST_FOREACH(StringProtoPair import, mImports)
    {
        import.second->InitialiseLibrary(reinit);
    }
    // Now do our library
    const unsigned library_size = mpLibrary->GetNumberOfDefinitions();
    assert(library_size == 0 || library_size == mLibraryStatements.size());
    if (library_size == 0)
    {
        mpLibrary->ExecuteStatements(mLibraryStatements);
    }
}


void Protocol::SetParalleliseLoops(bool paralleliseLoops)
{
    // The last clause checks that process isolation isn't already active
    mParalleliseLoops = paralleliseLoops && PetscTools::IsParallel() && !PetscTools::IsSequential();
}


void Protocol::Run()
{
    ProtocolTimer::BeginEvent(ProtocolTimer::RUN_PROTOCOL);
    ProtocolTimer::BeginEvent(ProtocolTimer::SETUP);
    std::cout << "Running protocol..." << std::endl;
    if (mpOutputHandler)
    {
        DebugProto::SetTraceFolder(*mpOutputHandler);
        mManifest.AddEntry("trace.txt", "text/plain");
    }
    ResetOutputs(mOutputs);
    // If we get an error at any stage, we want to ensure as many partial results as possible
    // are stored, but still report the error(s)
    std::vector<Exception> errors;
    ProtocolTimer::EndEvent(ProtocolTimer::SETUP);
    ProtocolTimer::BeginEvent(ProtocolTimer::SIMULATE);
    // Run the simulation(s)
    try
    {
        unsigned simulation_number = 0u;
        BOOST_FOREACH(boost::shared_ptr<AbstractSimulation> p_sim, mSimulations)
        {
            const std::string prefix = p_sim->GetOutputsPrefix();
            std::cout << "Running simulation " << simulation_number << " " << prefix
                      << " on process " << PetscTools::GetMyRank() << "..." << std::endl;
            p_sim->InitialiseSteppers();
            p_sim->SetParalleliseLoops(mParalleliseLoops);
            p_sim->Run();
            if (mpOutputHandler)
            {
                // Re-set the trace folder in case a nested protocol changed it
                DebugProto::SetTraceFolder(*mpOutputHandler);
                // Remove the simulation output folder if empty
                FileFinder sim_debug_output = p_sim->GetOutputFolder();
                if (sim_debug_output.IsPathSet() && sim_debug_output.IsDir() && sim_debug_output.IsEmpty() && PetscTools::AmMaster())
                {
                    sim_debug_output.Remove();
                }
            }
            simulation_number++;
        }
    }
    catch (const Exception& e)
    {
        std::cerr << e.GetMessage();
        errors.push_back(e);
        WriteError("Error running simulations:");
        WriteError(e);
    }
    ProtocolTimer::EndEvent(ProtocolTimer::SIMULATE);
    ProtocolTimer::BeginEvent(ProtocolTimer::POSTPROCESS);
    bool error_header_written = false;
    if (!mParalleliseLoops || PetscTools::AmMaster())
    {
        // Post-process the results
        EnvironmentPtr p_post_proc_env(new Environment(mpLibrary->GetAsDelegatee()));
        if (errors.empty())
        {
            std::cout << "Running post-processing..." << std::endl;
            try
            {
                p_post_proc_env->ExecuteStatements(mPostProcessing);
            }
            catch (const Exception& e)
            {
                std::cerr << e.GetMessage();
                errors.push_back(e);
                if (!error_header_written)
                {
                    WriteError("Error(s) running post-processing:");
                    error_header_written = true;
                }
                WriteError(e);
            }
        }
        // Transfer requested outputs to mOutputs[""]
        std::cout << "Recording protocol outputs..." << std::endl;
        EnvironmentPtr p_proto_outputs = mOutputs[""];
        BOOST_FOREACH(boost::shared_ptr<OutputSpecification> p_spec, mOutputSpecifications)
        {
            const std::string& r_loc = p_spec->GetLocationInfo();
            const std::string& r_ref = p_spec->rGetOutputRef();
            const std::string& r_name = p_spec->rGetOutputName();
            try
            {
                AbstractValuePtr p_output = p_post_proc_env->Lookup(r_ref, r_loc);
                if (p_spec->rGetOutputType() == "Post-processed")
                {
                    const std::string& r_units = p_spec->rGetOutputUnits();
                    if (!r_units.empty())
                    {
                        p_output->SetUnits(r_units);
                    }
                }
                p_proto_outputs->DefineName(r_name, p_output, r_loc);
            }
            catch (const Exception& e)
            {
                std::cerr << e.GetMessage();
                errors.push_back(e);
                WriteError(e);
            }
        }
    }
    if (mParalleliseLoops)
    {
        // Ensure all processes are in sync before the protocol finishes
        PetscTools::Barrier("Protocol::Run");
    }
    if (mpOutputHandler)
    {
        DebugProto::StopTracing();
        mManifest.WriteFile(*mpOutputHandler);
    }
    ProtocolTimer::EndEvent(ProtocolTimer::POSTPROCESS);
    ProtocolTimer::EndEvent(ProtocolTimer::RUN_PROTOCOL);
    PetscTools::ReplicateException(!errors.empty());
    if (!errors.empty())
    {
        THROW_REPORTED_EXCEPTIONS(errors);
    }
    std::cout << "Finished running protocol." << std::endl;
}


void Protocol::SetOutputFolder(const OutputFileHandler& rHandler)
{
    mpOutputHandler.reset(new OutputFileHandler(rHandler));
    BOOST_FOREACH(boost::shared_ptr<AbstractSimulation> p_sim, mSimulations)
    {
        p_sim->SetErrorFolder(mpOutputHandler);
        const std::string prefix = p_sim->GetOutputsPrefix();
        if (!prefix.empty())
        {
            FileFinder subfolder = rHandler.FindFile("simulation_" + prefix);
            boost::shared_ptr<OutputFileHandler> p_sim_handler(new OutputFileHandler(subfolder));
            p_sim->SetOutputFolder(p_sim_handler);
        }
    }
}


void Protocol::WriteToFile(const OutputFileHandler& rHandler,
                           const std::string& rFileNameBase)
{
    SetOutputFolder(rHandler);
    WriteToFile(rFileNameBase);
}


void Protocol::WriteError(const std::string& rMessage,
                          const OutputFileHandler& rHandler)
{
    try
    {
        out_stream p_error_file = rHandler.OpenOutputFile("errors.txt", std::ios_base::app);
        *p_error_file << rMessage << std::endl;
        p_error_file->close();
    }
    catch (...)
    {
        // Swallow any errors while writing - stdout will have full error information anyway.
    }
}

void Protocol::WriteError(const std::string& rMessage)
{
    if (!mpErrorHandler)
    {
        mpErrorHandler = mpOutputHandler;
    }
    if (mpErrorHandler)
    {
        Protocol::WriteError(rMessage, *mpErrorHandler);
        mManifest.AddEntry("errors.txt", "text/plain", true);
    }
}

void Protocol::WriteError(const Exception& rError)
{
    WriteError(ExceptionSet::ExtractShortMessage(rError));
}


void Protocol::SetErrorFolder(boost::shared_ptr<OutputFileHandler> pHandler)
{
    mpErrorHandler = pHandler;
    BOOST_FOREACH(boost::shared_ptr<AbstractSimulation> p_sim, mSimulations)
    {
        p_sim->SetErrorFolder(mpErrorHandler);
    }
}


void Protocol::RunAndWrite(const std::string fileNameBase)
{
    ProtocolTimer::BeginEvent(ProtocolTimer::ALL);
    // Record provenance info
    if (mpOutputHandler)
    {
        ExecutableSupport::SetOutputDirectory(mpOutputHandler->GetOutputDirectoryFullPath());
        ExecutableSupport::WriteMachineInfoFile("machine_info");
        ExecutableSupport::WriteProvenanceInfoFile();
        mManifest.AddEntry("machine_info_0.txt", "text/plain");
        mManifest.AddEntry("provenance_info_0.txt", "text/plain");
        mManifest.WriteFile(*mpOutputHandler);
    }
    // Run protocol
    try
    {
        Run();
    }
    catch (...)
    {
        if (mpOutputHandler)
        {
            try
            {
                std::cout << "Error running protocol. Trying to write intermediate results to file..." << std::endl;
                WriteToFile(fileNameBase);
                std::cout << "Intermediate results written; re-throwing error..." << std::endl;
            }
            catch (const Exception& e)
            {
                std::cout << "Failed to write intermediate results:" << std::endl << e.GetMessage();
            }
            catch (...)
            {
                std::cout << "Failed to write intermediate results." << std::endl;
            }
        }
        ProtocolTimer::EndEvent(ProtocolTimer::ALL);
        throw;
    }
    if (mpOutputHandler)
    {
        if (PetscTools::AmMaster())
        {
            std::cout << "Writing results to file..." << std::endl;
            WriteToFile(fileNameBase);
            // Indicate successful completion
            std::cout << "Done!" << std::endl;
            out_stream p_file = mpOutputHandler->OpenOutputFile("success");
            (*p_file) << "Protocol completed successfully." << std::endl;
            p_file->close();
            mManifest.AddEntry("success", "text/plain");
            mManifest.WriteFile(*mpOutputHandler);
        }
    }
    if (mParalleliseLoops)
    {
        // Ensure all processes are in sync before the protocol finishes
        PetscTools::Barrier("Protocol::RunAndWrite");
    }
    ProtocolTimer::EndEvent(ProtocolTimer::ALL);
}


std::string SanitiseFileName(const std::string& rFileName)
{
    std::string name = rFileName;
    FileFinder::ReplaceSpacesWithUnderscores(name);
    // Strip quote characters
    for (std::string::iterator it = name.begin(); it != name.end(); ++it)
    {
        while (*it == '\'' || *it == '"' || *it == '`'  || *it == '(' || *it == ')')
        {
            it = name.erase(it);
        }
    }
    return name;
}

std::string PlotFileName(boost::shared_ptr<PlotSpecification> pPlotSpec,
                         bool png=false)
{
    std::string ext;
    if (!png)
    {
        ext = ".eps";
    }
    else
    {
        ext = ".png";
    }
    std::string file_name = pPlotSpec->rGetTitle() + ext;
    return SanitiseFileName(file_name);
}

std::string DataFileName(const std::string& rFileNameBase, boost::shared_ptr<PlotSpecification> pPlotSpec)
{
    return SanitiseFileName(rFileNameBase + "_" + pPlotSpec->rGetTitle() + "_gnuplot_data.csv");
}


void Protocol::WriteToFile(const std::string& rFileNameBase)
{
    ProtocolTimer::BeginEvent(ProtocolTimer::WRITE_OUTPUTS);
    if (!mpOutputHandler)
    {
        EXCEPTION("SetOutputFolder must be called prior to using WriteToFile.");
    }
    if (!PetscTools::AmMaster())
    {
        return;
    }
    ///\todo Improve format?
    const Environment& r_outputs = rGetOutputsCollection();
    std::set<std::string> missing_outputs;
    // Variable metadata file
    {
        std::string index_file_name = rFileNameBase + "-contents.csv";
        out_stream p_file = mpOutputHandler->OpenOutputFile(index_file_name);
        mManifest.AddEntry(index_file_name, "text/csv");
        *p_file << "Variable id,Variable name,Units,Number of dimensions,File name,Type,Dimensions" << std::endl;
        BOOST_FOREACH(OutputSpecificationPtr p_spec, mOutputSpecifications)
        {
            const std::string& r_name = p_spec->rGetOutputName();
            AbstractValuePtr p_output;
            try
            {
                p_output = r_outputs.Lookup(r_name);
            }
            catch (const Exception&)
            {
                std::cerr << "Missing protocol output '" << r_name << "'; ignoring for now." << std::endl;
                missing_outputs.insert(r_name);
                continue;
            }
            if (p_output->IsArray())
            {
                NdArray<double> output = GET_ARRAY(p_output);
                *p_file << '"' << r_name << "\",\"" << p_spec->rGetOutputDescription()
                         << "\",\"" << p_output->GetUnits()
                         << "\"," << output.GetNumDimensions()
                         << "," << SanitiseFileName(rFileNameBase + "_" + r_name + ".csv")
                         << "," << p_spec->rGetOutputType();
                BOOST_FOREACH(NdArray<double>::Index len, output.GetShape())
                {
                    *p_file << "," << len;
                }
                *p_file << std::endl;
            }
            else
            {
                std::cerr << "Unexpected non-array protocol output '" << r_name << "'; ignoring." << std::endl;
            }
        }
    }

    // Default plots file
    if (!mPlotSpecifications.empty())
    {
        std::string file_name = rFileNameBase + "-default-plots.csv";
        out_stream p_file = mpOutputHandler->OpenOutputFile(file_name);
        mManifest.AddEntry(file_name, "text/csv");
        *p_file << "Plot title,File name,Data file name,Line style,First variable id,Optional second variable id,Optional key variable id" << std::endl;
        BOOST_FOREACH(PlotSpecificationPtr p_spec, mPlotSpecifications)
        {
            bool all_variables_exist = true;
            std::vector<std::string> units;
            std::vector<std::string> descriptions;
            // Check all the variables are also outputs, or we can't plot them
            BOOST_FOREACH(const std::string& r_name, p_spec->rGetVariableNames())
            {
                try
                {
                    AbstractValuePtr p_output = r_outputs.Lookup(r_name);
                    units.push_back(p_output->GetUnits());
                    // Look up the output description for this plot
                    BOOST_FOREACH(OutputSpecificationPtr p_output_spec, mOutputSpecifications)
                    {
                        if (r_name == p_output_spec->rGetOutputName())
                        {
                            descriptions.push_back(p_output_spec->rGetOutputDescription());
                            break;
                        }
                    }
                }
                catch (const Exception&)
                {
                    std::cerr << "Plot requests protocol output '" << r_name << "', which has not been specified as an output." << std::endl;
                    missing_outputs.insert(r_name);
                    all_variables_exist = false;
                }
            }
            if (all_variables_exist)
            {
                p_spec->SetVariableUnits(units); // Store for later plotting use
                p_spec->SetVariableDescriptions(descriptions); // Store for later plotting use
                const std::string& r_title = p_spec->rGetTitle();
                *p_file << '"' << r_title << "\"," << PlotFileName(p_spec);
                *p_file << "," << DataFileName(rFileNameBase, p_spec) << "," << p_spec->rGetStyle();
                BOOST_FOREACH(const std::string& r_name, p_spec->rGetVariableNames())
                {
                    *p_file << ",\"" << r_name << "\"";
                }
                if (p_spec->rGetVariableNames().size() == 1u)
                {
                    // Second variable omitted, so add a blank entry
                    *p_file << ",";
                }
                if (!p_spec->rGetKeyVariableName().empty())
                {
                    *p_file << ",\"" << p_spec->rGetKeyVariableName() << "\"";
                }
                *p_file << std::endl;
            }
        }
    }

    // Output data, one file per variable
    BOOST_FOREACH(OutputSpecificationPtr p_spec, mOutputSpecifications)
    {
        const std::string& r_name = p_spec->rGetOutputName();
        AbstractValuePtr p_output;
        try
        {
            p_output = r_outputs.Lookup(r_name);
        }
        catch (const Exception&)
        {
            continue;
        }

        std::string file_name = SanitiseFileName(rFileNameBase + "_" + r_name + ".csv");
        out_stream p_file = mpOutputHandler->OpenOutputFile(file_name);
        mManifest.AddEntry(file_name, "text/csv");
        *p_file << "# " << p_spec->rGetOutputDescription() << std::endl;

        if (p_output->IsArray())
        {
            NdArray<double> output = GET_ARRAY(p_output);
            const unsigned num_dims = output.GetNumDimensions();
            NdArray<double>::Extents shape = output.GetShape();
            if (num_dims == 2)
            {
                // Tabular format for easy processing by other software
                //*p_file << std::scientific;
                *p_file << std::setprecision(16);
                NdArray<double>::Indices idxs = output.GetIndices();
                for (idxs[1]=0; idxs[1]<shape[1]; ++idxs[1])
                {
                    for (idxs[0]=0; idxs[0]<shape[0]; ++idxs[0])
                    {
                        *p_file << (idxs[0] == 0 ? "" : ",") << output[idxs];
                    }
                    *p_file << std::endl;
                }
            }
            else
            {
                // General format - one entry per line, with header giving the shape
                *p_file << num_dims;
                for (unsigned i=0; i<num_dims; ++i)
                {
                    *p_file << "," << shape[i];
                }
                *p_file << std::endl;
                //*p_file << std::scientific;
                *p_file << std::setprecision(16);
                for (NdArray<double>::Iterator it = output.Begin(); it != output.End(); ++it)
                {
                    *p_file << *it << std::endl;
                }
            }
        }
        p_file->close();
    }
    mManifest.WriteFile(*mpOutputHandler);
    ProtocolTimer::EndEvent(ProtocolTimer::WRITE_OUTPUTS);

    GeneratePlots(rFileNameBase);

    if (!missing_outputs.empty())
    {
        std::vector<std::string> missing_names(missing_outputs.begin(), missing_outputs.end());
        WriteError("Not all expected protocol outputs were defined; see full output for details.");
        EXCEPTION("Not all protocol outputs were defined.  Missing names: " << missing_names);
    }
}


void Protocol::SetPngOutput(bool writePng)
{
    mWritePng = writePng;
}


/**
 * Helper function to add units to an axis label if they're useful.
 * We ignore blank and dimensionless units.
 *
 * @param label  the base label
 * @param rUnits  the units name/description
 * @return  the augmented label
 */
std::string GetAxisLabel(std::string label, const std::string& rUnits)
{
    if (rUnits != "" && rUnits != "dimensionless")
    {
        label += " (" + rUnits + ")";
    }
    return label;
}


void Protocol::GeneratePlots(const std::string& rFileNameBase)
{
    ProtocolTimer::BeginEvent(ProtocolTimer::WRITE_PLOTS);
    const Environment& r_outputs = rGetOutputsCollection();

    // \todo #1999 LOADS OF TESTS!!
    BOOST_FOREACH(boost::shared_ptr<PlotSpecification> p_plot_spec, mPlotSpecifications)
    {
        const std::string& r_title = p_plot_spec->rGetTitle();
        const std::vector<std::string>& r_names = p_plot_spec->rGetVariableNames();

        // Get the arrays to plot
        std::vector<AbstractValuePtr> arrays;
        bool missing_array = false;
        BOOST_FOREACH(const std::string& r_name, r_names)
        {
            try
            {
                AbstractValuePtr p_output = r_outputs.Lookup(r_name);
                if (p_output->IsArray())
                {
                    arrays.push_back(p_output);
                }
                else
                {
                    std::cerr << "Cannot plot non-array output '" << r_name << "'." << std::endl;
                    missing_array = true;
                }
            }
            catch (const Exception&)
            {
                missing_array = true;
                // Error for missing output was already printed by caller
            }
        }
        // Check that the key vector is present as well, if specified
        if (!p_plot_spec->rGetKeyVariableName().empty())
        {
            const std::string& r_name = p_plot_spec->rGetKeyVariableName();
            try
            {
                AbstractValuePtr p_key = r_outputs.Lookup(r_name);
                if (!p_key->IsArray() || !GET_ARRAY(p_key).GetNumDimensions() == 1u)
                {
                    std::cerr << "Cannot plot non-vector key variable '" << r_name << "'." << std::endl;
                    missing_array = true;
                }
            }
            catch (const Exception&)
            {
                missing_array = true;
                std::cerr << "Plot requests key vector '" << r_name << "' which is not specified as an output." << std::endl;
            }
        }
        // Don't plot if any required output is missing
        if (missing_array)
        {
            continue;
        }

        // Figure out axis labels, and the X axis data array
        std::string label_x, label_y;
        NdArray<double> output_x;
        if (r_names.size() == 1u)
        {
            // Find the output specification for this variable
            BOOST_FOREACH(OutputSpecificationPtr p_output_spec, mOutputSpecifications)
            {
                if (r_names.front() == p_output_spec->rGetOutputName())
                {
                    PROTO_ASSERT2(p_output_spec->rGetOutputType() == "Raw",
                                  "Only 'Raw' type outputs may be plotted using the 'data' form.",
                                  p_plot_spec->GetLocationInfo());
                    // Find the simulation it comes from
                    const std::string& r_ref = p_output_spec->rGetOutputRef();
                    size_t colon = r_ref.find(':');
                    PROTO_ASSERT2(colon != std::string::npos,
                                  "A raw output reference must be a prefixed name, not '" << r_ref << "'.",
                                  p_output_spec->GetLocationInfo());
                    std::string prefix = r_ref.substr(0, colon);
                    BOOST_FOREACH(AbstractSimulationPtr p_sim, mSimulations)
                    {
                        if (prefix == p_sim->GetOutputsPrefix())
                        {
                            std::vector<AbstractStepperPtr>& r_steppers = p_sim->rGetSteppers();
                            if (!r_steppers.empty() && r_steppers.back())
                            {
                                // Create a new array containing the stepper data
                                AbstractStepperPtr p_stepper = r_steppers.back();
                                PROTO_ASSERT2(p_stepper->IsEndFixed(),
                                              "Unable to plot against a while loop at present.",
                                              p_plot_spec->GetLocationInfo());
                                label_x = GetAxisLabel(p_stepper->GetIndexName(), p_stepper->GetUnits());
                                NdArray<double>::Extents x_shape(1u);
                                x_shape[0] = p_stepper->GetNumberOfOutputPoints();
                                NdArray<double> x(x_shape);
                                output_x = x;
                                NdArray<double>::Iterator it_x = x.Begin();
                                for (p_stepper->Reset(); !p_stepper->AtEnd(); p_stepper->Step())
                                {
                                    *it_x++ = p_stepper->GetCurrentOutputPoint();
                                }
                            }
                            break;
                        }
                    }
                    break;
                }
            }
            PROTO_ASSERT2(!label_x.empty(), "Unable to find X axis label.", p_plot_spec->GetLocationInfo());
            PROTO_ASSERT2(output_x.GetNumElements() > 0u, "Unable to find X axis data.",
                          p_plot_spec->GetLocationInfo());
            label_y = GetAxisLabel(p_plot_spec->rGetVariableDescriptions()[0], p_plot_spec->rGetVariableUnits()[0]);
        }
        else
        {
            label_x = GetAxisLabel(p_plot_spec->rGetVariableDescriptions()[0], p_plot_spec->rGetVariableUnits()[0]);
            if (r_names.size() == 2u)
            {
                label_y = GetAxisLabel(p_plot_spec->rGetVariableDescriptions()[1], p_plot_spec->rGetVariableUnits()[1]);
            }
            else
            {
                label_y = "Various";
            }
            output_x = GET_ARRAY(arrays[0]);
            arrays.erase(arrays.begin());
        }

        // Tracing what we're doing
        std::cout << "Plotting " << r_title << ":\t" << label_y << " against " << label_x << "..." << std::endl;

        // Check the shapes of the arrays to plot
        // X must be 1d, or be equivalent to a 1d array (i.e. just multiple copies of the same vector)
        const unsigned x_length = output_x.GetShape().back();
        if (output_x.GetNumDimensions() != 1)
        {
            // Create a 1d slice of X
            std::vector<NdArray<double>::Range> ranges;
            for (unsigned dim=0; dim<output_x.GetNumDimensions()-1; ++dim)
            {
                ranges.push_back(NdArray<double>::Range(0u));
            }
            ranges.push_back(NdArray<double>::Range(0u, NdArray<double>::Range::END));
            NdArray<double> output_x_1d = output_x[ranges];
            // Compare each possible slice to it
            bool equal = true;
            NdArray<double>::Indices idxs_1d = output_x_1d.GetIndices();
            for (NdArray<double>::ConstIterator it = output_x.Begin(); it != output_x.End(); ++it)
            {
                idxs_1d.front() = it.rGetIndices().back();
                if (*it != output_x_1d[idxs_1d])
                {
                    equal = false;
                    break;
                }
            }
            if (!equal)
            {
                std::cerr << "The X data for a plot must be a 1d array, not " << output_x.GetNumDimensions() << "d." << std::endl;
                continue;
            }
            else
            {
                output_x = output_x_1d;
            }
        }
        // Y arrays can be n-d, but the last dimension must match the length of X
        std::vector<NdArray<double> > y_arrays;
        std::vector<unsigned> num_traces;
        unsigned total_num_traces = 0u;
        BOOST_FOREACH(AbstractValuePtr p_array_val, arrays)
        {
            NdArray<double> output_y = GET_ARRAY(p_array_val);
            y_arrays.push_back(output_y);
            if (output_y.GetNumDimensions() < 1)
            {
                std::cerr << "The Y data for a plot must be a true array, not a single value." << std::endl;
                continue;
            }
            if (output_y.GetShape().back() != x_length)
            {
                std::cerr << "The last dimension of the Y data for a plot must be the same size as the X data. "
                          << "Y shape " << output_y.GetShape() << " is not compatible with X length " << x_length << std::endl;
                continue;
            }
            // Each Y array may represent multiple traces if it has more than one dimension
            unsigned my_num_traces = 1u;
            for (unsigned i=0; i<output_y.GetNumDimensions()-1; ++i)
            {
                my_num_traces *= output_y.GetShape()[i];
            }
            total_num_traces += my_num_traces;
            num_traces.push_back(my_num_traces);
        }

        if (total_num_traces == 0u)
        {
            std::cerr << "No valid traces found." << std::endl;
            continue;
        }

        // Write data for plotting
        std::string file_name = DataFileName(rFileNameBase, p_plot_spec);
        out_stream p_file = mpOutputHandler->OpenOutputFile(file_name);
        mManifest.AddEntry(file_name, "text/csv");
        // Tabular format with no header line for easy processing by gnuplot
        for (NdArray<double>::ConstIterator it_x=output_x.Begin();
             it_x != output_x.End();
             ++it_x)
        {
            (*p_file) << *it_x;
            // Now we iterate over Y data
            for (unsigned i=0; i<y_arrays.size(); ++i)
            {
                NdArray<double>& output_y = y_arrays[i];
                // Find the first Y entry for this X position
                NdArray<double>::ConstIterator it_y = output_y.Begin();
                it_y += it_x.rGetIndices().back();
                for (unsigned trace=0; trace<num_traces[i]; ++trace)
                {
                    (*p_file) << "," << *it_y;
                    // Increment by the size of the last dimension to point at the next trace for this X
                    if (trace < num_traces[i]-1)
                    {
                        it_y += output_y.GetShape().back();
                    }
                }
            }
            (*p_file) << std::endl;
        }
        p_file->close();

        // Plot!
        PlotWithGnuplot(p_plot_spec, file_name, total_num_traces, x_length, label_x, label_y, false);
        if (mWritePng)
        {
            PlotWithGnuplot(p_plot_spec, file_name, total_num_traces, x_length, label_x, label_y, true);
        }
    }
    mManifest.WriteFile(*mpOutputHandler);
    ProtocolTimer::EndEvent(ProtocolTimer::WRITE_PLOTS);
}


/**
 * Escape single quotes in the given string, by doubling.
 *
 * @param rString
 */
void EscapeQuotes(std::string& rString)
{
    for (std::string::iterator it = rString.begin(); it != rString.end(); ++it)
    {
        if (*it == '\'')
        {
            it = rString.insert(it, '\'') + 1;
        }
    }
}


void Protocol::PlotWithGnuplot(PlotSpecificationPtr pPlotSpec,
                               const std::string& rDataFileName,
                               const unsigned numTraces,
                               const unsigned numPointsInTrace,
                               std::string xLabel,
                               std::string yLabel,
                               bool writePng)
{
    /// \todo #1999 generalise further
    /// \todo #1999 make the plot title include the model as well as protocol name

    // Find the csv file name, and remove .csv to make the .gp file name
    size_t dot = rDataFileName.rfind('.');
    std::string script_name = rDataFileName.substr(0, dot) + ".gp";
    std::string fig_file_name = PlotFileName(pPlotSpec, writePng);

    // Decide whether to plot with points or lines (default points joined by straight lines)
    std::string points_or_lines = pPlotSpec->rGetStyle();
    assert(!points_or_lines.empty());

    // Escape single-quote characters in the plot title & axis labels
    std::string display_title = pPlotSpec->rGetDisplayTitle();
    EscapeQuotes(display_title);
    EscapeQuotes(xLabel);
    EscapeQuotes(yLabel);

    // Write out a Gnuplot script
    out_stream p_gnuplot_script = mpOutputHandler->OpenOutputFile(script_name);
    mManifest.AddEntry(script_name, "text/plain", true);
    std::string output_dir = mpOutputHandler->GetOutputDirectoryFullPath();
    *p_gnuplot_script << "# Gnuplot script file generated by Chaste Functional Curation system." << std::endl;
    *p_gnuplot_script << "# Plot of " << pPlotSpec->rGetTitle() << "." << std::endl;
    *p_gnuplot_script << "set terminal ";
    if (!pPlotSpec->rGetGnuplotTerminal().empty())
    {
        // Page format was specified manually
        *p_gnuplot_script << pPlotSpec->rGetGnuplotTerminal();
    }
    else if (!writePng)
    {
        // eps page size in inches, default font
        *p_gnuplot_script << "postscript eps enhanced size 3,2 font 16";
    }
    else
    {
        // png size in pixels, font requires something like:
        // export GDFONTPATH=/usr/share/fonts/truetype/ttf-liberation/
        *p_gnuplot_script << "png enhanced size 900,600 font \"LiberationSans-Regular\" 16";
    }
    *p_gnuplot_script << std::endl;
    *p_gnuplot_script << "set output \"" << output_dir << fig_file_name << "\"" << std::endl;
    *p_gnuplot_script << "set title '" << display_title << "'" << std::endl;
    *p_gnuplot_script << "set xlabel '" << xLabel << "'" << std::endl;
    *p_gnuplot_script << "set ylabel '" << yLabel << "'" << std::endl;
    *p_gnuplot_script << "set grid" << std::endl;
    *p_gnuplot_script << "set autoscale" << std::endl;
    *p_gnuplot_script << "set datafile separator \",\"" << std::endl;

    // Add a key if requested
    const std::string& r_key_name = pPlotSpec->rGetKeyVariableName();
    std::vector<double> key_values;
    if (r_key_name.empty())
    {
        *p_gnuplot_script << "set key off" << std::endl;
    }
    else
    {
        AbstractValuePtr p_key = rGetOutputsCollection().Lookup(r_key_name);
        // Look up the output description for the key to get its description
        std::string key_desc;
        BOOST_FOREACH(OutputSpecificationPtr p_output_spec, mOutputSpecifications)
        {
            if (r_key_name == p_output_spec->rGetOutputName())
            {
                key_desc = p_output_spec->rGetOutputDescription();
                break;
            }
        }
        std::string key_title = GetAxisLabel(key_desc, p_key->GetUnits());
        *p_gnuplot_script << "set key outside invert title \"" + key_title + "\"" << std::endl;
        // Get the key values for use in setting curve titles below
        assert(p_key->IsArray());
        NdArray<double> key = GET_ARRAY(p_key);
        key_values.reserve(key.GetNumElements());
        for (NdArray<double>::ConstIterator iter = key.Begin(); iter != key.End(); ++iter)
        {
            key_values.push_back(*iter);
        }
        if (key_values.size() != numTraces)
        {
            std::cerr << "Plot key vector '" << r_key_name << "' has " << key_values.size()
                      << " entries, but there are " << numTraces << " traces." << std::endl;
            key_values.clear();
        }
    }

    // Any special-case setup
    BOOST_FOREACH(const std::string& rCommand, pPlotSpec->rGetGnuplotExtraCommands())
    {
        *p_gnuplot_script << rCommand << std::endl;
    }

    // The actual plot command...
    *p_gnuplot_script  << "plot ";
    for (unsigned i=0; i<numTraces; ++i)
    {
        *p_gnuplot_script << "\"" << output_dir << rDataFileName << "\" using 1:" << i+2;
        if (!key_values.empty())
        {
            *p_gnuplot_script << " title \"" << key_values[i] << "\"";
        }
        *p_gnuplot_script << " with " << points_or_lines;
        if (i < numTraces-1)
        {
            *p_gnuplot_script << ",\\"; // Escape the newline that is written below
        }
        *p_gnuplot_script << std::endl;
    }
    *p_gnuplot_script << std::endl << std::flush;

    p_gnuplot_script->close();

    // Run Gnuplot on the script written above to generate image files.
    int ret = system(("gnuplot " + output_dir + script_name).c_str());
    if (ret != 0)
    {
        WARNING("Unable to generate plot '" << pPlotSpec->rGetTitle() << "' due to missing Gnuplot.");
    }
    else
    {
        if (writePng)
        {
            mManifest.AddEntry(fig_file_name, "image/png");
        }
        else
        {
            mManifest.AddEntry(fig_file_name, "application/postscript");
        }
    }
}


unsigned Protocol::GetNumberOfOutputs(unsigned simulation) const
{
    if (simulation == static_cast<unsigned>(-1))
    {
        simulation = mSimulations.size() - 1;
    }
    EXCEPT_IF_NOT(simulation < mSimulations.size());
    EXCEPT_IF_NOT(!mSimulations[simulation]->GetOutputsPrefix().empty());
    return GetNumberOfOutputs(mSimulations[simulation]->GetOutputsPrefix());
}


const Environment& Protocol::rGetOutputsCollection(unsigned simulation) const
{
    if (simulation == static_cast<unsigned>(-1))
    {
        simulation = mSimulations.size() - 1;
    }
    EXCEPT_IF_NOT(simulation < mSimulations.size());
    EXCEPT_IF_NOT(!mSimulations[simulation]->GetOutputsPrefix().empty());
    return rGetOutputsCollection(mSimulations[simulation]->GetOutputsPrefix());
}


unsigned Protocol::GetNumberOfOutputs(const std::string& rPrefix) const
{
    std::map<std::string, EnvironmentPtr>::const_iterator it = mOutputs.find(rPrefix);
    EXCEPT_IF_NOT(it != mOutputs.end());
    return it->second->GetNumberOfDefinitions();
}


const Environment& Protocol::rGetOutputsCollection(const std::string& rPrefix) const
{
    std::map<std::string, EnvironmentPtr>::const_iterator it = mOutputs.find(rPrefix);
    EXCEPT_IF_NOT(it != mOutputs.end());
    return *(it->second);
}


boost::shared_ptr<ModelStateCollection> Protocol::GetStateCollection()
{
    return mpModelStateCollection;
}


void Protocol::SetInput(const std::string& rName, AbstractExpressionPtr pValue)
{
    mpInputs->RemoveDefinition(rName, "Setting protocol input");
    AbstractValuePtr p_value = (*pValue)(*mpInputs);
    mpInputs->DefineName(rName, p_value, "Setting protocol input");
}


Manifest& Protocol::rGetManifest()
{
    return mManifest;
}


const FileFinder& Protocol::rGetSourceFile() const
{
    return mSourceFilePath;
}


ProtocolPtr Protocol::GetImportedProtocol(const std::string& rPrefix)
{
    return mImports[rPrefix];
}


const std::map<std::string, ProtocolPtr> Protocol::rGetImportedProtocols() const
{
    return mImports;
}


const std::map<std::string, std::string>& Protocol::rGetNamespaceBindings() const
{
    return mOntologyNamespaceBindings;
}


Environment& Protocol::rGetInputsEnvironment()
{
    return *mpInputs;
}


std::vector<AbstractStatementPtr>& Protocol::rGetInputStatements()
{
    return mInputStatements;
}


Environment& Protocol::rGetLibrary()
{
    return *mpLibrary;
}


std::vector<AbstractStatementPtr>& Protocol::rGetLibraryStatements()
{
    return mLibraryStatements;
}


std::vector<boost::shared_ptr<AbstractSimulation> >& Protocol::rGetSimulations()
{
    return mSimulations;
}


std::vector<AbstractStatementPtr>& Protocol::rGetPostProcessing()
{
    return mPostProcessing;
}


std::vector<OutputSpecificationPtr>& Protocol::rGetOutputSpecifications()
{
    return mOutputSpecifications;
}


std::vector<PlotSpecificationPtr>& Protocol::rGetPlotSpecifications()
{
    return mPlotSpecifications;
}


void Protocol::SetSourceFile(const FileFinder& rSourceFile)
{
    mSourceFilePath = rSourceFile;
}


void Protocol::AddImport(const std::string& rPrefix, ProtocolPtr pImport, const std::string& rLoc)
{
    std::pair<std::map<std::string, ProtocolPtr>::iterator, bool> result = mImports.insert(std::make_pair(rPrefix, pImport));
    PROTO_ASSERT2(result.second,
                  "The prefix " << rPrefix << " has already been used for an imported protocol.",
                  rLoc);
    // Make the import's library, if any, available to our library and hence its users
    mpLibrary->SetDelegateeEnvironment(pImport->rGetLibrary().GetAsDelegatee(), rPrefix);
}


void Protocol::AddNamespaceBindings(const std::map<std::string, std::string>& rBindings)
{
    BOOST_FOREACH(StringPair new_binding, rBindings)
    {
        std::pair<std::map<std::string, std::string>::iterator,bool> result = mOntologyNamespaceBindings.insert(new_binding);
        if (!result.second)
        {
            // Binding of this prefix already existed.  Check it used the same URI.
            PROTO_ASSERT2(result.first->second == new_binding.second,
                          "Conflicting definitions of namespace prefix " << new_binding.first << ": was "
                          << result.first->second << " but tried to bind to " << new_binding.second,
                          "Protocol definition");
        }
    }
}


void Protocol::AddInputDefinitions(const std::vector<AbstractStatementPtr>& rStatements)
{
    std::copy(rStatements.begin(), rStatements.end(), std::back_inserter(mInputStatements));
}


void Protocol::AddPostProcessing(const std::vector<AbstractStatementPtr>& rStatements)
{
    mPostProcessing.resize(mPostProcessing.size() + rStatements.size());
    std::copy(rStatements.begin(), rStatements.end(), mPostProcessing.end()-rStatements.size());
}


void Protocol::AddLibrary(const std::vector<AbstractStatementPtr>& rStatements)
{
    mLibraryStatements.resize(mLibraryStatements.size() + rStatements.size());
    std::copy(rStatements.begin(), rStatements.end(), mLibraryStatements.end()-rStatements.size());
}


void Protocol::AddSimulation(boost::shared_ptr<AbstractSimulation> pSimulation)
{
    mSimulations.push_back(pSimulation);
}


void Protocol::AddSimulations(const std::vector<boost::shared_ptr<AbstractSimulation> >& rSimulations)
{
    std::copy(rSimulations.begin(), rSimulations.end(), std::back_inserter(mSimulations));
}


void Protocol::AddOutputSpecs(const std::vector<boost::shared_ptr<OutputSpecification> >& rSpecifications)
{
    mOutputSpecifications.resize(mOutputSpecifications.size() + rSpecifications.size());
    std::copy(rSpecifications.begin(), rSpecifications.end(), mOutputSpecifications.end()-rSpecifications.size());
}


void Protocol::AddDefaultPlots(const std::vector<boost::shared_ptr<PlotSpecification> >& rSpecifications)
{
    mPlotSpecifications.resize(mPlotSpecifications.size() + rSpecifications.size());
    std::copy(rSpecifications.begin(), rSpecifications.end(), mPlotSpecifications.end()-rSpecifications.size());
}


//
// Associating a protocol with a model
//

void Protocol::CheckModel(boost::shared_ptr<AbstractSystemWithOutputs> pModel) const
{
    if (!pModel)
    {
        EXCEPTION("Given model doesn't have any protocol outputs.");
    }
    if (pModel->GetNumberOfOutputs() == 0)
    {
        EXCEPTION("Number of model outputs is zero.");
    }
}


void Protocol::SetModel(boost::shared_ptr<AbstractSystemWithOutputs> pModel)
{
    CheckModel(pModel);
    mpModel = pModel;
    // Set up the model wrapper environments
    mpModel->SetNamespaceBindings(mOntologyNamespaceBindings);
    // Associate each simulation with this model
    // (and hence set up delegation to the model via the simulation)
    BOOST_FOREACH(AbstractSimulationPtr p_sim, mSimulations)
    {
        p_sim->SetModel(mpModel);
    }
    // Now run all the library programs, if present
    InitialiseLibrary();
}


boost::shared_ptr<AbstractSystemWithOutputs> Protocol::GetModel()
{
    EXCEPT_IF_NOT(mpModel);
    return mpModel;
}
