/*

Copyright (c) 2005-2014, University of Oxford.
All rights reserved.

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
 * Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 * Neither the name of the University of Oxford nor the names of its
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#ifndef FUNCTIONCALL_HPP_
#define FUNCTIONCALL_HPP_

#include "AbstractExpression.hpp"

#include <string>
#include <vector>

/**
 * Expression that calls a function when evaluated.
 *
 * It takes an expression that evaluates to the function (often a NameLookup) and expressions
 * to evaluate to provide the function arguments.
 *
 * Default values for arguments can be requested by using DefaultParameter as an actual argument.
 */
class FunctionCall : public AbstractExpression
{
public:
    /**
     * Create a function call.
     *
     * @param pFunction  the expression giving the function to call
     * @param rParameters  the expressions giving the function arguments
     */
    FunctionCall(const AbstractExpressionPtr pFunction, const std::vector<AbstractExpressionPtr>& rParameters);

    /**
     * Convenience constructor for the common case where the function is looked up in the
     * environment.
     *
     * @param rName  the name of the function to call
     * @param rParameters  the expressions giving the function arguments
     */
    FunctionCall(const std::string& rName, const std::vector<AbstractExpressionPtr>& rParameters);

    /**
     * Evaluate the expression in an environment.
     *
     * Note that this method is still virtual here - built-in operations inherit from FunctionCall.
     * \todo Consider whether it makes more sense for them to inherit from LambdaClosure and be defined in
     * every environment?
     *
     * @param rEnv  the environment
     */
    virtual AbstractValuePtr operator()(const Environment& rEnv) const;

private:
    /** The expression giving the function to call. */
    AbstractExpressionPtr mpFunction;
};

#endif // FUNCTIONCALL_HPP_
